from scipy.interpolate import CubicSpline  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import ArrayLike, NDArray
from numpy import float64, linspace, array

from science_signal import interpolate


class Signal:
    """
    A signal with amplitude linked to times/frequencies
    """

    def __init__(
        self, x: NDArray[float64], y: NDArray[float64]
    ) -> None:
        if len(x) != len(y):
            raise ValueError(
                _(
                    "each value of a signal must be linked with a "
                    + "frequency/time"
                )
            )
        self.x = x
        self.y = y
        self.index = -1
        try:
            self.sampling: float = self.x[1] - self.x[0]
        except IndexError:
            raise ValueError(
                _("the signal must contains more than one value")
            )
        try:
            self.rate = 1 / self.sampling
        except ZeroDivisionError:
            raise ValueError(
                _(
                    "each value of the signal should be linked with a "
                    + "single frequency/time"
                )
            )

    def spline(self) -> CubicSpline:
        """
        Return the x to y spline of this signal
        """
        return CubicSpline(self.x, self.y)

    def operator_signal(
        self, other: "Signal", operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal with another one
        """
        try:
            if (self.x != other.x).any():  # fail if different length
                raise ValueError(
                    "not an error"
                )  # fail if different value
        except ValueError:
            self, other = interpolate(self, other)
        if operator == "+":
            return Signal(
                self.x,
                self.y + other.y,
            )
        if operator == "-":
            return Signal(self.x, self.y - other.y)
        if operator == "*":
            return Signal(
                self.x,
                self.y * other.y,
            )
        if operator == "/":
            return Signal(
                self.x,
                self.y / other.y,
            )

        raise Exception(_("unknown operator {}").format(operator))

    def operator_arraylike(
        self, other: ArrayLike, operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal (signal operator other)
        """
        try:
            if len(self) != len(other):  # fail if single value
                # suppose same range but different rates
                return self.operator_signal(
                    Signal(
                        linspace(self.x[0], self.x[-1], len(others)),
                        array(other),
                    ),
                    operator,
                )
        except TypeError:
            pass
        if operator == "+":
            return Signal(
                self.x,
                self.y + other,
            )
        if operator == "-":
            return Signal(
                self.x,
                self.y - other,
            )
        if operator == "*":
            return Signal(
                self.x,
                self.y * other,
            )
        if operator == "/":
            return Signal(
                self.x,
                self.y / other,
            )
        raise Exception(_("unkown operator {}".format(operator)))

    def roperator_arraylike(
        self, other: ArrayLike, operator: str
    ) -> "Signal":
        """
        Apply an operator to this signal (other operator signal)
        """
        try:
            if len(self) != len(other):  # fail if single value
                return Signal(
                    linspace(self.x[0], self[-1], len(other)),
                    array(other),
                ).operator_signal(self, operator)
        except TypeError:
            pass
        if operator == "-":
            return Signal(
                self.x,
                other - self.y,
            )
        if operator == "/":
            return Signal(
                self.x,
                other / self.y,
            )
        raise Exception(_("unknown operator {}".format(operator)))

    def __add__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "+")
        return self.operator_arraylike(other, "+")

    def __rad__(self, other: ArrayLike | "Signal") -> "Signal":
        return self + other  # symetric

    def __sub__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "-")
        return self.operator_arraylike(other, "-")

    def __rsub__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return other.operator_signal(self, "-")
        return self.roperator_arraylike(other, "-")

    def __mul__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "*")
        return self.operator_arraylike(other, "*")

    def __rmul__(self, other: ArrayLike | "Signal") -> "Signal":
        return self * other

    def __div__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return self.operator_signal(other, "/")
        return self.operator_arraylike(other, "/")

    def __rdiv__(self, other: ArrayLike | "Signal") -> "Signal":
        if isinstance(other, Signal):
            return other.operator_signal(self, "/")
        return self.roperator_arraylike(other, "/")

    def __array__(self) -> NDArray[float64]:
        return array([self.x, self.y])

    def __abs__(self) -> "Signal":
        return Signal(
            self.x,
            abs(self.y),
        )

    def __neg__(self) -> "Signal":
        return Signal(
            self.x,
            -self.y,
        )

    def __len__(self) -> int:
        return len(self.x)

    def __next__(self) -> tuple[float, float]:
        self.index += 1
        if self.index == len(self):
            raise StopIteration
        return self.x[self.index], self.y[self.index]

    def __getitem__(self, key: int) -> tuple[float, float]:
        return self.x[key], self.y[key]

    def __iter__(self) -> "Signal":
        self.index = -1
        return self
